#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <stdint.h>
#include <unistd.h>
#include <sys/mman.h>
#include <dlfcn.h>
#include "NetSock.h"
#include "../common.h"
#include <pthread.h>
#include <sys/ptrace.h>


static volatile bool g_finished;
static NetSock *g_net;

static api_t g_api;

static unsigned char g_chain_value;

typedef unsigned char (*master_func_t)(void *api, unsigned char);

void api_ack(unsigned char v) {
  g_net->WriteAll(&v, 1);
}

void api_send_to_master(const void *data, int sz) {
  g_net->WriteAll(data, sz);
}

int api_recv_from_master(void *buffer, int sz) {
  return g_net->ReadAll(buffer, sz);
}

void api_set_finished() {
  g_finished = true;
}

void api_flush_stdout() {
  fflush(stdout);
}

void *api_mmap(unsigned long length, int prot) {
  length = (length + 4095) & 0xfffUL;
  return mmap(NULL, length, prot, MAP_ANON | MAP_PRIVATE, -1, 0);
}

void api_msleep(int ms) {
  usleep(ms * 1000);
}

int api_thread_create(void *thread, void *(*start_routine)(void *), void *arg) {
  return pthread_create((pthread_t*)thread, NULL, start_routine, arg);
}

void api_gets(char *dst, int sz) {
  memset(dst, 0, sz);
  for (int i = 0; i < sz - 1; i++) {
    int ch = getchar();
    if (ch == '\n') {
      break;
    }
    dst[i] = ch;
  }
}

void *api_dlsym(void *handle, const char *symbol) {
  return dlsym(handle, symbol);
}

void init_api() {
  g_api.ack = api_ack;
  g_api.send_to_master = api_send_to_master;
  g_api.recv_from_master = api_recv_from_master;
  g_api.set_finished = api_set_finished;
  g_api.puts = puts;
  g_api.printf = printf;
  g_api.flush_stdout = api_flush_stdout;
  g_api.abort = _exit;
  g_api.mmap = api_mmap;
  g_api.msleep = api_msleep;
  g_api.gets = api_gets;
  g_api.thread_create = api_thread_create;
  g_api.dlsym = api_dlsym;
  g_api.dlopen = dlopen;
  g_api.dlclose = dlclose;
}

int run(const char *master_host) {
  init_api();

  puts("Welcome to Chains.");
  printf("Connecting to master..."); fflush(stdout);

  NetSock net;
  if (!net.Connect(master_host, 7679)) {
    puts("FAILED");
    return 1;
  }
  puts("OK");

  net.ReadAll(&g_chain_value, 1);

  g_net = &net;
  while (!g_finished) {
    int code_size = 0;
    int ret = net.ReadAll(&code_size, sizeof(code_size));
    if (ret != sizeof(code_size)) {
      fprintf(stderr, "Master disconnected (timeout?)\n");
      return 1;
    }

    const size_t padded_code_size = ((size_t)code_size + 4095) & (~0xfffULL);
    void *code = mmap(NULL, padded_code_size,
                      PROT_EXEC | PROT_READ | PROT_WRITE,
                      MAP_ANON | MAP_PRIVATE, -1, 0);

    if (code == nullptr) {
      fprintf(stderr, "mmap failed, this should not happen unless you're doing sth weird.\n");
      return 2;
    }

    ret = net.ReadAll(code, code_size);
    if (ret != code_size) {
      fprintf(stderr, "Master disconnected (timeout?)\n");
      return 3;
    }

    master_func_t func = (master_func_t)code;
    g_chain_value = func(&g_api, g_chain_value);
    munmap(code, padded_code_size);
  }

  return 0;
}

int main(void) {
  const char *master_host = "chains.of.trust.felinae.pl";
  const char *alt_master_host = getenv("MASTER_HOST");
  if (alt_master_host != nullptr) {
    master_host = alt_master_host;
  }

  return run(master_host);
}

