#!/usr/bin/python
import sys
import threading
import os
import pickle
import random
import types
from common import *

import helper
import challenge

HERE = os.path.dirname(os.path.abspath(__file__))

g_debug = False

g_active_thread_count = 0
g_functions = None
g_helper_functions = {}
g_challenge_functions = {}

g_proof_of_work = [
    "proof_of_work1",
    "proof_of_work2",
    "proof_of_work3",
    "proof_of_work4",
    "proof_of_work5",
]

g_anti_dbg = [
  # "ptrace" is actually called separately from this.
  "ptrace2",  # This one knows that ptrace has been called.
  "env1",
  "env2",
  "parent1",
  "libs1",
  # "libs2" is called separately as it's CPU/IO intensive.
  "exec1",  # Similar to libs2, but the file is smaller.
]

g_stages = [
    (1, [ "precheck" ]),
    (10, g_proof_of_work ),
    (1, [ "ptrace" ] ),  # This can be called only once.
    (10, g_anti_dbg + g_proof_of_work),
    (1, [ "libs2" ] ),  # CPU intensive, don't call it too often.
    (1, [ "display" ]),  # Init io device.
    (10, g_anti_dbg + g_proof_of_work + [ "thread_noise" ]),
    (4, [ "ram" ]),   # Init 4 RAM devices.
    (10, g_anti_dbg + g_proof_of_work),
    (1, [ "libs2" ] ),  # CPU intensive, don't call it too often.
    (4, [ "apu" ]),   # Init 4 Auxiliary Processing Units.
    (10, g_anti_dbg + g_proof_of_work),
    (1, [ "cpu1" ]),  # Init Stage 1 CPU.
    (10, g_anti_dbg + g_proof_of_work),
    (1, [ "cpu2" ]),  # Init Stage 2 CPU.
    (10, g_anti_dbg + g_proof_of_work + [ "thread_noise" ]),
    (1, [ "cpu3" ]),  # Init Stage 3 CPU.
    (10, g_anti_dbg + g_proof_of_work),

    # Wait 12 * 5 seconds.
    (1, [ "end" ]), (3, g_anti_dbg + g_proof_of_work + [ "thread_noise" ]),
    (1, [ "end" ]), (3, g_anti_dbg + g_proof_of_work),
    (1, [ "end" ]), (3, g_anti_dbg + g_proof_of_work),
    (1, [ "end" ]), (3, g_anti_dbg + g_proof_of_work + [ "thread_noise" ]),
    (1, [ "end" ]), (3, g_anti_dbg + g_proof_of_work),
    (1, [ "end" ]), (3, g_anti_dbg + g_proof_of_work),
    (1, [ "end" ]), (3, g_anti_dbg + g_proof_of_work),
    (1, [ "end" ]), (3, g_anti_dbg + g_proof_of_work + [ "thread_noise" ]),
    (1, [ "end" ]), (3, g_anti_dbg + g_proof_of_work),
    (1, [ "end" ]), (3, g_anti_dbg + g_proof_of_work),
    (1, [ "end" ]), (3, g_anti_dbg + g_proof_of_work + [ "thread_noise" ]),
    (1, [ "end" ]), (3, g_anti_dbg + g_proof_of_work),

    (1, [ "hard_end" ])
]

def fill_dict_with_functions_from_module(mod, d):
  for el_name in dir(mod):
    if el_name.startswith("_"):
      continue

    el = getattr(mod, el_name)
    if type(el) is not types.FunctionType:
      continue

    d[el_name] = el

def init_functions():
  global g_functions
  with open("%s/chunks.pickle" % HERE, "rb") as f:
    g_functions = pickle.load(f)

  fill_dict_with_functions_from_module(helper, g_helper_functions)
  fill_dict_with_functions_from_module(challenge, g_challenge_functions)


class HandlerThread(threading.Thread):
  def __init__(self, s):
    threading.Thread.__init__(self)
    self.s = s
    self.ptr_storage = {}

  def send_code(self, chunk_id):
    assert chunk_id in g_functions
    code = random.choice(g_functions[chunk_id])
    code_size = dd(len(code))
    self.s.sendall(code_size)
    self.s.sendall(code)

    func = g_helper_functions.get(chunk_id)
    if func is not None:
      return func(self, self.s, chunk_id)

    return True

  def worker(self):
    self.done = False

    self.chain_value = random.randint(0, 255)
    self.s.sendall(db(self.chain_value))

    if g_debug:
      print "--- Chain Value %i" % self.chain_value

    stage = 0
    for repeat, chunks in g_stages:
      if g_debug:
        print "--v Stage %i" % stage

      for i in xrange(repeat):
        chunk = random.choice(chunks)

        if g_debug:
          print "  + Chunk %s" % chunk

        # Send code.
        res = self.send_code(chunk)
        if res is False:
          return

        # Iterate chain of trust.
        challenge = random.randint(0, 255)
        self.s.sendall(db(challenge))

        challenge_func = g_challenge_functions.get(chunk)
        assert challenge_func is not None

        self.chain_value = challenge_func(self.chain_value, challenge)
        assert 0 <= self.chain_value <= 255

        # Wait for reply.
        data = recvall(self.s, 1)
        if data is False or len(data) != 1:
          return

        if rb(data[0]) != self.chain_value:
          self.send_code("broken_chain")
          if g_debug:
            print "  + Broken Chain: %i vs %i" % (rb(data[0]), self.chain_value)
          return

      stage += 1

    return

  def run(self):
    global g_active_thread_count
    g_active_thread_count += 1

    self.s.settimeout(10)  # 10 seconds max.
    sys.stdout.write("%i " % g_active_thread_count)
    sys.stdout.flush()

    try:
      self.worker()
    except socket.timeout as e:
      pass
    except socket.error as e:
      if e.errno == 107 or e.errno == 32:  # Client disconnected.
        pass
      else:
        print "Socket (W)", e


    try:
      self.s.shutdown(socket.SHUT_WR)
    except socket.error as e:
      if e.errno == 107:  # Client already disconnected.
        pass
      else:
        print "Socket", e

    self.s.close()
    g_active_thread_count -= 1
    sys.stdout.write("%i " % g_active_thread_count)
    sys.stdout.flush()

def main():
  print "Initializing..."
  init_functions()

  s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
  s.bind(("0.0.0.0", 7679))
  s.listen(123)

  print "Serving!"

  while True:
    conn, addr = s.accept()
    #print addr, "(connections active: %i)" % g_active_thread_count
    h = HandlerThread(conn)
    h.daemon = True
    h.start()


if __name__ == "__main__":
  sys.exit(main())


