#include "common_ch.h"
#include "../../common.h"

__attribute__((section(".start")))
unsigned char main(struct api_t *api, unsigned char chain_value) {
  s64 (*readlink)(const char*, char*, size_t) = api->dlsym(RTLD_DEFAULT, "readlink");
  void* (*fopen)(const char*, const char*) = api->dlsym(RTLD_DEFAULT, "fopen");
  int (*fclose)(void*) = api->dlsym(RTLD_DEFAULT, "fclose");
  size_t (*fread)(void*, size_t, size_t, void*) = api->dlsym(RTLD_DEFAULT, "fread");
  int (*sprintf)(char*, const char*, ...) = api->dlsym(RTLD_DEFAULT, "sprintf");
  void* (*memset)(void*, int, size_t) = api->dlsym(RTLD_DEFAULT, "memset");

  char exe_path[256] = {0};
  if (readlink("/proc/self/exe", exe_path, 255) == -1) {
    api->puts("I need /proc/ access. Go away.");
    api->set_finished();
    return 0;
  }

  // Replace last / with \0.
  char *p = exe_path + 255;
  while (p >= exe_path) {
    if (*p == '/') {
      *p = '\0';
      break;
    }
    p--;
  }

  struct hash_st {
    const char *name;
    u64 hash;
  } h[] = {
    { "libc.so.6", 878103062796601758ULL },
    { "libpthread.so.0", 17446952141597939731ULL },
    { "ld-linux-x86-64.so.2", 177558958204707110ULL },
    { NULL, 0 }
  };

  for (int i = 0; h[i].name != NULL; i++) {
    char what[512] = {0};
    sprintf(what, "%s/%s", exe_path, h[i].name);

    void *f = fopen(what, "rb");
    if (!f) {
      api->puts("You are doing something weird. Go away.");
      api->set_finished();
      return 0;
    }

    u64 hash = 0x123456789abcdef0ULL;

    char buf[1024 * 64];
    memset(buf, 0, 1024 * 64);
    for (;;) {
      size_t ret = fread(buf, 1, 1024 * 64, f);
      if (ret == 0) {
        break;
      }

      for (size_t j = 0; j < ret; j += 8) {
        u64 *v = (u64*)(buf + j);
        hash ^= *v;
      }
    }

    fclose(f);

    if (hash != h[i].hash) {
      api->puts("No debugging, go away.");
      api->set_finished();
      return 0;
    }
  }

  // Usual challenge stuff otherwise.
  unsigned char challenge;
  if (api->recv_from_master(&challenge, 1) != 1) {
    api->puts("Master disconnected.");
    api->set_finished();
    return 0;
  }

  unsigned char new_chain_value = (challenge ^ 0xff ^ chain_value) & 0xff;
  api->ack(new_chain_value);
  return new_chain_value;
}

