#include "../../common.h"

#define NULL 0
#define RTLD_LAZY 1

void worker(struct api_t *api) {
  api->puts("Checking required libraries...");

  const char *libs[] = {
    "libc.so.6",
    "libpthread.so.0",
    NULL
  };

  for (const char **l = libs; *l; l++) {
    api->printf("  + %s\n", *l);
    void *ret = api->dlopen(*l, RTLD_LAZY);
    if (ret == NULL) {
      api->printf("Missing required library: %s\n", *l);
      api->flush_stdout();
      api->abort(1);
    }
  }

  api->puts("Looks good so far, but will do some more checks.");
  api->puts("This might take a moment...");
}

__attribute__((section(".start")))
unsigned char main(struct api_t *api, unsigned char chain_value) {
  worker(api);

  unsigned char challenge;
  if (api->recv_from_master(&challenge, 1) != 1) {
    api->puts("Master disconnected.");
    api->set_finished();
    return 0;
  }

  unsigned char new_chain_value = (chain_value * 17 + challenge) & 0xff;
  api->ack(new_chain_value);
  return new_chain_value;
}

