#pragma once

#define PROT_READ  1
#define PROT_WRITE 2
#define PROT_EXEC  4
#define PROT_RWX   7
#define PROT_RW    3

#define NULL 0

#define RTLD_LAZY 1
#define RTLD_NOW 2
#define RTLD_GLOBAL 0x100
#define RTLD_DEFAULT ((void*)NULL)

typedef unsigned long size_t;
typedef unsigned long int pthread_t;

typedef unsigned char bool;
#define true 1
#define false 0

typedef unsigned long  u64;
typedef unsigned int   u32;
typedef unsigned short u16;
typedef unsigned char  u8;

typedef long         s64;
typedef int          s32;
typedef short        s16;
typedef signed char  s8;

// Virtual Machine structures follow.

struct api_t;

struct vm_io_dev_st {
  struct api_t *api;
  bool power_off;

  bool input_ready;
  bool output_ready;

  char input_buffer[64];
  char output_buffer[256];
};

struct ram_port_st {
  u16 addr;
  u16 data;
  bool addr_ready;  // Triggers read or write. RAM will reset it.
  bool data_ready;  // Set by RAM, indicated data ready on read.
                    // Client has to reset it on its own.
  bool write_enabled;  // When set, data will be written.
};

struct vm_ram_dev_st {
  struct api_t *api;
  bool power_off;

  struct ram_port_st port[4];  // APU uses port 0, CPU's use ports 1-3.
  u16 data[128];
};

// RAM layout (in words):
// 0x00 - 0x07   APU input/output
// 0x20          APU enabled flag (set by CPU1)
// 0x21          APU finished flag (set by APU)
// -- The rest is only in bank 3.
// 0x40 - 0x5F   CPU2 gathered and process data.

struct vm_apu_dev_st {
  struct api_t *api;
  bool power_off;

  u8 id;
  struct ram_port_st *ram_port;
};

struct vm_cpu1_dev_st {
  struct api_t *api;
  bool power_off;

  bool cpu2_ready;  // Set by CPU2 when both CPU2 and CPU3 are ready.
  bool apu_finished;  // Set by CPU1 when APUs are done.

  struct vm_io_dev_st *io;
  struct ram_port_st *ram_port[4];
};

struct vm_cpu2_dev_st {
  struct api_t *api;
  bool power_off;

  bool cpu3_ready;  // Set by CPU3 when it's ready.
  bool data_ready;  // Set by CPU2 for CPU3 to continue.

  struct vm_cpu1_dev_st *cpu1;
  struct ram_port_st *ram_port[4];
};

struct vm_cpu3_dev_st {
  struct api_t *api;
  bool power_off;

  struct vm_io_dev_st *io;
  struct vm_cpu2_dev_st *cpu2;
  struct ram_port_st *ram_port[4];
};

// So technically this shouldn't be needed, but actually I don't want
// to waste too much time debugging weird crashes later in that one
// obscure place where I forgot sth, so I'll add these here and there
// anyway.
#define mfence() asm volatile ("mfence" ::: "memory")
