from common import *
import random

# Helper functions (for proof of work, etc).

# Devices.
def display(self, s, chunk_id):
  io_dev_ptr = recvall(s, 8)
  if io_dev_ptr is False or len(io_dev_ptr) != 8:
    return False

  self.ptr_storage["io"] = io_dev_ptr
  return True

def ram(self, s, chunk_id):
  io_ram_ptr = recvall(s, 8)
  if io_ram_ptr is False or len(io_ram_ptr) != 8:
    return False

  if "ram" not in self.ptr_storage:
    self.ptr_storage["ram"] = [ io_ram_ptr ]
  else:
    self.ptr_storage["ram"].append(io_ram_ptr)
  return True

def apu(self, s, chunk_id):
  io_apu_ptr = recvall(s, 8)
  if io_apu_ptr is False or len(io_apu_ptr) != 8:
    return False

  if "apu" not in self.ptr_storage:
    apu_id = 0
    self.ptr_storage["apu"] = [ io_apu_ptr ]
  else:
    apu_id = len(self.ptr_storage["apu"])
    self.ptr_storage["apu"].append(io_apu_ptr)

  s.sendall(db(apu_id))
  s.sendall(self.ptr_storage["ram"][apu_id])
  return True

def cpu1(self, s, chunk_id):
  io_cpu1_ptr = recvall(s, 8)
  if io_cpu1_ptr is False or len(io_cpu1_ptr) != 8:
    return False

  self.ptr_storage["cpu1"] = io_cpu1_ptr

  s.sendall(self.ptr_storage["io"])
  s.sendall(''.join(self.ptr_storage["ram"]))  # Send all RAM pointers.

  return True

def cpu2(self, s, chunk_id):
  io_cpu2_ptr = recvall(s, 8)
  if io_cpu2_ptr is False or len(io_cpu2_ptr) != 8:
    return False

  self.ptr_storage["cpu2"] = io_cpu2_ptr

  s.sendall(self.ptr_storage["cpu1"])
  s.sendall(''.join(self.ptr_storage["ram"]))  # Send all RAM pointers.
  return True

def cpu3(self, s, chunk_id):
  io_cpu3_ptr = recvall(s, 8)
  if io_cpu3_ptr is False or len(io_cpu3_ptr) != 8:
    return False

  self.ptr_storage["cpu3"] = io_cpu3_ptr

  s.sendall(self.ptr_storage["io"])
  s.sendall(self.ptr_storage["cpu2"])
  s.sendall(''.join(self.ptr_storage["ram"]))  # Send all RAM pointers.
  return True

# End.
def end(self, s, chunk_id):
  s.sendall(self.ptr_storage["io"])
  return True

def hard_end(self, s, chunk_id):
  return False  # Kill the connection.


# Proof of work.
def _handle_proof_of_work(s, work, good_res):
  magic = recvall(s, 4)
  if magic is False or magic != "WoRK":
    return False

  s.sendall(work)
  res = recvall(s, 1)

  if res is False or len(res) != 1:
    return False

  res = rb(res)
  return res == good_res

def proof_of_work1(self, s, chunk_id):
  work = dd(random.randint(0, 0xffffffff))
  good_res = sum([rb(x) for x in work]) & 0xff
  return _handle_proof_of_work(s, work, good_res)

def proof_of_work2(self, s, chunk_id):
  work = dd(random.randint(0, 0xffffffff))
  good_res = (rd(work) ^ 0x9876bd45) & 0xff
  return _handle_proof_of_work(s, work, good_res)

def proof_of_work3(self, s, chunk_id):
  work = dd(random.randint(0, 0xffffffff))
  good_res = (rd(work) + 0x17823bcd) & 0xff
  return _handle_proof_of_work(s, work, good_res)

def proof_of_work4(self, s, chunk_id):
  work = dd(random.randint(0, 0xffffffff))
  good_res = (rd(work) * 0x881bfffd) & 0xff
  return _handle_proof_of_work(s, work, good_res)

def proof_of_work5(self, s, chunk_id):
  work = dd(random.randint(0, 0xffffffff))
  good_res = (rd(work) - 0xaaabbbcc) & 0xff
  return _handle_proof_of_work(s, work, good_res)

# Other.
def template(self, s, chunk_id):
  magic = recvall(s, 4)
  if magic is False or magic != "WoRK":
    return False

  work = "1234"

  s.sendall(work)
  res = recvall(s, 1)

  if res is False or len(res) != 1:
    return False

  res = ord(res)
  good_res = sum([ord(x) for x in work]) & 0xff
  return res == good_res

